﻿using System.Collections;
using System.Collections.Generic;
using Photon.Pun;
using UnityEngine;
 
public class Ball : MonoBehaviour
{
    public Color saturatedColor;
    private bool edible = false;
    private bool onFloor = false;
    private const float NOT_ON_FLOOR_EDIBLE_DURATION = 2f;
    private float notOnFloorFloorEdibleTimeLeft = NOT_ON_FLOOR_EDIBLE_DURATION;
    private float originalDiameter = 1f;
    


    public virtual void Awake()
    {
        originalDiameter = Random.Range(Game.MIN_BALL_DIAMETER, Game.MAX_BALL_DIAMETER);
        transform.localScale = new Vector3(originalDiameter, originalDiameter, originalDiameter);
        saturatedColor = GetRandomColor();
        GetComponent<Renderer>().material.color = CalculateDesaturatedColor(saturatedColor);

        Physics.IgnoreLayerCollision(0, 8);

        if (Random.Range(0, 10) == 0) // was 5
        {
            Mesh randomMesh = Resources.Load<GameObject>("Prefabs/ball01mesh").GetComponent<MeshFilter>().sharedMesh;
            GetComponent<MeshFilter>().mesh = randomMesh;
        }
    }


    public static Color GetRandomColor()
    {
        Color randomColor;
        const int NUM_COLORS = 7;

        int randomColorIndex = Random.Range(0, NUM_COLORS);
        if (randomColorIndex == 0)
        {
            randomColor = Color.red;
        }
        else if (randomColorIndex == 1)
        {
            randomColor = Color.blue;
        }
        else if (randomColorIndex == 2)
        {
            randomColor = Color.cyan;
        }
        else if (randomColorIndex == 3)
        {
            randomColor = Color.yellow;
        }
        else if (randomColorIndex == 4)
        {
            randomColor = Color.Lerp(Color.yellow, Color.red, 0.5f); //orange
        }
        else if (randomColorIndex == 5)
        {
            randomColor = Color.Lerp(Color.cyan, Color.blue, 0.5f); //purple
        }
        else //(randomColorIndex == 6)
        {
            randomColor = Color.Lerp(Color.cyan, Color.blue, 0.5f); //purple
        }

        return randomColor;
    }

    // Start is called before the first frame update
    void Start()
    {

    }

    // Update is called once per frame
    void Update()
    {
        if (IsEdible())
        {
            if (!onFloor)
            {
                notOnFloorFloorEdibleTimeLeft -= Time.deltaTime;

                if (notOnFloorFloorEdibleTimeLeft < 0f)
                {
                    SetEdible(false);
                }
            }

            //pingpong color value between black and white to obtain a pulsing effect
            GetComponent<Renderer>().material.color = Color.Lerp(Color.black, Color.white, Mathf.PingPong(Time.time * 2f, 1f));
        }


    }

    public void DisablePhysics()
    {
        GetComponent<Collider>().enabled = false;
        GetComponent<Rigidbody>().isKinematic = true;
    }

    public void SetEdible(bool edible)
    {
        if (edible && !onFloor)
        {
            Debug.LogWarning("A ball that is not on the floor has been set to be edible! GameObject " + transform.name);
        }

        this.edible = edible;

        if (edible)
        {
            gameObject.GetComponent<Renderer>().material.EnableKeyword("_EMISSION");
            gameObject.GetComponent<Renderer>().material.SetColor("_EmissionColor", Color.gray);
        }
        else
        {
            gameObject.GetComponent<Renderer>().material.DisableKeyword("_EMISSION");
            GetComponent<Renderer>().material.color = CalculateDesaturatedColor(saturatedColor);
        }
    }

    public void SetEdibleVisuals(bool edible)
    {
        if (edible)
        {
            gameObject.GetComponent<Renderer>().material.EnableKeyword("_EMISSION");
            gameObject.GetComponent<Renderer>().material.SetColor("_EmissionColor", Color.gray);
        }
        else
        {
            gameObject.GetComponent<Renderer>().material.DisableKeyword("_EMISSION");
            GetComponent<Renderer>().material.color = CalculateDesaturatedColor(saturatedColor);
        }
    }

    public bool IsEdible()
    {
        return edible;
    }

    public void SetOnFloor(bool onFloor)
    {
        if (onFloor)
        {
            notOnFloorFloorEdibleTimeLeft = NOT_ON_FLOOR_EDIBLE_DURATION;
        }

        this.onFloor = onFloor;
    }

    public bool IsOnFloor()
    {
        return onFloor;
    }

    public float GetOriginalDiameter()
    {
        return originalDiameter;
    }

    public void EnableSnakeColor(bool enabled)
    {
        if(enabled)
        {
            GameObject snakeBody = Resources.Load<GameObject>("Prefabs/snakeBall01mesh");
            GetComponent<MeshFilter>().mesh = snakeBody.GetComponent<MeshFilter>().sharedMesh;
            GetComponent<Renderer>().material = snakeBody.GetComponent<Renderer>().sharedMaterial;
            transform.rotation = Quaternion.identity;
           // GetComponent<Renderer>().material.color = Color.Lerp(Color.Lerp(Color.black, Color.green, 0.5f), saturatedColor, 0.5f);
        }
        else
        {
            GetComponent<Renderer>().material.color = CalculateDesaturatedColor(saturatedColor);
        }
    }

    private Color CalculateDesaturatedColor(Color saturatedColor)
    {
        return Color.Lerp(Color.white, saturatedColor, 0.5f);
    }

}
