﻿using System.Collections;
using System.Collections.Generic;
using ExitGames.Client.Photon;
using Photon.Pun;
using Photon.Realtime;
using UnityEngine;

[RequireComponent( typeof( Ball ), typeof(PhotonView))]
public class BallPhoton : MonoBehaviourPun, IPunObservable
{
    private bool isUpdateRequired = false;
    private Vector3 hostPosition = Vector3.zero;
    private Vector3 hostScale = Vector3.zero;
    private bool requestDestroyed = false;
    

    public virtual void Awake()
    {
        if (!photonView.IsMine)
        {
            GetComponent<Rigidbody>().isKinematic = true;
        }
    }


    public void OnPhotonSerializeView(PhotonStream stream, PhotonMessageInfo info)
    {
        if (stream.IsWriting)
        {
            //We own this player: send the others our data
            stream.SendNext(requestDestroyed);
            stream.SendNext(transform.position);
            stream.SendNext(transform.localScale);
            stream.SendNext(GetComponent<Ball>().IsEdible());
            stream.SendNext(
                GameObject.FindObjectOfType<Snake>().HasBall( GetComponent<Ball>() )
            );
        }
        else
        {
            //Network player, receive data
            bool hostDestroyed = (bool)stream.ReceiveNext();
            if (hostDestroyed)
            {
                Debug.LogError("Host Destroyed");
                Destroy(gameObject);
                isUpdateRequired = false;
                return;
            }
            isUpdateRequired = true;
            hostPosition = (Vector3)stream.ReceiveNext();
            hostScale = (Vector3)stream.ReceiveNext();
            bool isEdible = (bool)stream.ReceiveNext();
            GetComponent<Ball>().SetEdibleVisuals(isEdible);
            bool isSnakeBall = (bool)stream.ReceiveNext();
            GetComponent<Ball>().EnableSnakeColor(isSnakeBall);
        }
    }

    // Start is called before the first frame update
    void Start()
    {

    }

    // Update is called once per frame
    void Update()
    {
        if (!photonView.IsMine && isUpdateRequired)
        {
            //Update Object position and Rigidbody parameters
            transform.position = Vector3.Lerp(transform.position, hostPosition, Time.deltaTime * 5);
            transform.localScale = Vector3.Lerp(transform.localScale, hostScale, Time.deltaTime * 5);
        }
    }


    private void OnDestroy()
    {
        requestDestroyed = true;
		if (PhotonNetwork.IsMasterClient)
		{
			PhotonNetwork.Destroy(gameObject);
		}
    }
}