﻿using System.Collections;
using System.Collections.Generic;
using Photon.Pun;
using UnityEngine;

[RequireComponent( typeof( Crane ), typeof( PhotonView ) )]
public class CranePhoton : MonoBehaviourPun, IPunObservable
{

    private Vector3 hostPositionClaw;
    private bool isUpdateRequired = false;
    private Crane crane;


    public virtual void Awake()
    {
        crane = GetComponent<Crane>();
    }



    [PunRPC]
    private void SetTargetPosition(Vector3 position)
    {
        crane.SetTargetInputFromServer(position);
    }


    public void OnPhotonSerializeView(PhotonStream stream, PhotonMessageInfo info)
    {
        if (stream.IsWriting)
        {
            // This is the client calling the shots
            stream.SendNext(crane.transformClaw.position);
            stream.SendNext(crane.GetState());
        }
        else
        {
            //Network player, receive data
            hostPositionClaw = (Vector3)stream.ReceiveNext();
            Crane.State hostCraneState = (Crane.State)stream.ReceiveNext();

            if (hostCraneState == Crane.State.OPEN_CLAW)
            {
                crane.OpenClawDirectly();
            }
            else if (hostCraneState == Crane.State.CLOSE_CLAW)
            {
                crane.CloseClawDirectly();
            }

            isUpdateRequired = true;
        }
    }


    // Update is called once per frame
    void Update()
    {
        if (!photonView.IsMine && isUpdateRequired)
        {
            crane.transformClaw.position = Vector3.Lerp(crane.transformClaw.position, hostPositionClaw, Time.deltaTime * 5);
        }
    }
}