
using ExitGames.Client.Photon;
using Photon.Pun;
using Photon.Pun.UtilityScripts;
using Photon.Realtime;
using UnityEngine;
using UnityEngine.UI;
using System.Collections.Generic;

public class Game : MonoBehaviourPunCallbacks
{
    public const int GAME_VERSION = 1;
   

    //tuning
    public const float BALL_PIT_WIDTH = 2.35f;
    public const float BALL_PIT_DEPTH = 1.27f;
    public const float BALL_PIT_WALL_HEIGHT = 2f;
    public const float BALL_PIT_FLOOR_THICKNESS = 1.0f;
    public const float BALL_PIT_WALL_THICKNESS = 0.25f;
    public const float MIN_BALL_DIAMETER = 0.075f * 2f * 1.125f;
    public const float MAX_BALL_DIAMETER = 0.085f * 2f * 1.125f;
    //TODO verify that this is a good amount of balls
    public const int MIN_BALL_AMOUNT = 150;
    public const int MAX_BALL_AMOUNT = MIN_BALL_AMOUNT; //min and max should remain the same
    public const float SNAKE_HEAD_DIAMETER = 0.125f * 1.25f * 1.125f;
    public const int NUM_EDIBLES = 5;

    public const int NUM_COINS_MAX = 1500; // 15 dollars
    public const int NUM_COINS_START = 1000; // 10 dollars
    public const int NUM_COINS_PER_MOVE = 25; // 25 cents
    public const int NUM_COINS_LOOT = 5; // 5 cents
    public const float GAME_OVER_MESSAGE_DURATION = 7.5f;


    //other
    private State state = State.INITIAL;
    private PlayerType playerType = PlayerType.INITIAL;
    public float gameOverMessageTimeLeft = GAME_OVER_MESSAGE_DURATION;
    string gameOverMessage = "DEFAULT GAME OVER MESSAGE";

    private GameObject craneCamera;
    private GameObject snakeCamera;
    private GameObject titleCamera;

    private int numBallsEatenBySnake = 0;

    private int numCoinsCurrent = NUM_COINS_START;

    //TODO Bernard verify this is all used in the calcuation for the progress bar
    private int numBallsGrabbedByCrane = 0;
    private int numBallsDestroyedByLevel = 0;


    private enum PlayerType
    {
        INITIAL,
        CRANE,
        SNAKE
    }

    private enum State
    {
        INITIAL,
        SELECTING_PLAYMODE,
        SHOWING_TITLE_SCREEN,
        INITIALIZE_NETWORK,
        JOIN_OR_START_ROOM,
        WAITING_FOR_PLAYER_TWO,
        LOAD_GAME,
        TIMER,
        PLAYING,
        SHOWING_GAME_OVER_MESSAGE,
    }

    // ui
    public Text networkInfo;

    // Networking
    public const int MAX_PLAYERS = 2;
    public const string PLAYER_READY = "IsPlayerReady";
    public const string PLAYER_LOADED_LEVEL = "PlayerLoadedLevel";

    private bool isConnectdToMaster = false;

    public const string PROP_KEY_P1 = "P1";
    private TypedLobby sqlLobby = new TypedLobby("customSqlLobby", LobbyType.SqlLobby);
    private bool isRoomlistUptodate = false;
    private List<RoomInfo> roomList = new List<RoomInfo>();


    private enum PlayMode
    {
        Unselected,
        Multiplayer,
        LocalDebug,
    }
    private PlayMode playMode = PlayMode.Unselected;
    private PlayerType playerTypeSelected = PlayerType.INITIAL;
    private double timerStartTime = 0;

    private void Awake()
    {
    }

    void Start()
    {
        ChangeState(State.SELECTING_PLAYMODE);
    }


    void Update()
    {
        if (state == State.SELECTING_PLAYMODE)
        {
            if (playMode != PlayMode.Unselected)
            {
                ChangeState(State.SHOWING_TITLE_SCREEN);
            }
        }
        else if (state == State.SHOWING_TITLE_SCREEN)
        {
            if (playerTypeSelected != PlayerType.INITIAL)
            {
                ChangePlayerType(playerTypeSelected);
                if (playMode == PlayMode.LocalDebug)
                {
                    ChangeState(State.LOAD_GAME);
                }
                else if (playMode == PlayMode.Multiplayer)
                {
                    ChangeState(State.INITIALIZE_NETWORK);
                }
            }
        }
        else if (state == State.INITIALIZE_NETWORK)
        {
            if (IsNetworkInitialized())
            {
                ChangeState(State.JOIN_OR_START_ROOM);
            }
        }
        else if (state == State.JOIN_OR_START_ROOM)
        {
            if (isRoomlistUptodate)
            {
                RoomInfo roomToJoin = FindRoom();
                if (roomToJoin != null)
                {
                    networkInfo.text = "Found opponent '" + roomToJoin.Name + "'; get ready..";
                    PhotonNetwork.JoinRoom(roomToJoin.Name);
                    ChangeState(State.WAITING_FOR_PLAYER_TWO);
                }
                else
                {
                    // Cranes wait for Snakes to make rooms
                    if (playerType == PlayerType.SNAKE)
                    {
                        networkInfo.text = "Creating new game, waiting for opponent to join..";
                        CreateRoom();
                        ChangeState(State.WAITING_FOR_PLAYER_TWO);
                    }
                    else
                    {
                        GetCustomRoomList("" + (int)GetPlayerTypeOpponent(playerType));
                    }
                }
            }
        }
        else if (state == State.WAITING_FOR_PLAYER_TWO)
        {
            if (PhotonNetwork.CurrentRoom != null && PhotonNetwork.CurrentRoom.PlayerCount > 1)
            {
                ChangeState(State.LOAD_GAME);
            }
        }
        else if (state == State.LOAD_GAME)
        {
            ChangeState(State.TIMER);
        }
        else if (state == State.TIMER)
        {
            double timer = CalculateElapsedTime();
            if (timer > 4.0)
            {
                networkInfo.text = "3";
            }

            else if (timer > 3.0)
            {
                networkInfo.text = "GO!";
                ChangeState(State.PLAYING);
            }

            else if (timer > 2.0)
            {
                networkInfo.text = "1";
            }

            else if (timer > 1.0)
            {
                networkInfo.text = "2";
            }
            //networkInfo.text = "" + timer + "\nStt:" + timerStartTime;
        }
        else if (state == State.PLAYING)
        {
            // cleans up any balls that have escaped our leaky colliders
            DestroyFallenBalls();

            // set targets / edible balls
            if (playerType == PlayerType.SNAKE)
            {
                HideBallsNotOnFloor();
                SetBallsEdible();
            }

            // have snake eat balls
            Snake snake = GameObject.FindObjectOfType<Snake>();
            Ball lastEdibleBallNearMouth = snake.GetLastEdibleBallNearMouth();
            if (lastEdibleBallNearMouth)
            {
                //the order of the following two calls is important as they both set the color; by calling EatBall last the balls that belong to the snake can look different
                lastEdibleBallNearMouth.SetEdible(false);
                if (snake.EatBall(lastEdibleBallNearMouth))
                {
                    numBallsEatenBySnake++;
                }

            }

            Crane crane = GameObject.FindObjectOfType<Crane>();
            GameObject craneLoot = crane.GetAndRemoveLootObject();
            if (craneLoot != null)
            {
                if (craneLoot.gameObject == snake.gameObject)
                {
                    // Crane won!

                }
                else
                {
                    Debug.Log("Crane loot destroyed");
                    snake.SafelyDestroyBall(craneLoot.GetComponent<Ball>());

                    numBallsGrabbedByCrane++;
                    numCoinsCurrent += NUM_COINS_LOOT;
                }

            }
            numCoinsCurrent -= crane.GetAndResetNumMoves() * NUM_COINS_PER_MOVE;

            //TODO remove after the Split functionality has been fully tested
            // shortcut: S splits the snake on the 4th position (needs 4 or more balls attached to the snake)
            /*
            if (Input.GetKeyDown(KeyCode.S))
            {
                snake.Split(3, null);
            }
            */

            // shortcut: E ends the game (for now); consider removing
            if (TestGameOverConditionAndSetMessage() || Input.GetKeyDown(KeyCode.E))
            {
                ChangeState(State.SHOWING_GAME_OVER_MESSAGE);
            }
        }
        else if (state == State.SHOWING_GAME_OVER_MESSAGE)
        {
            gameOverMessageTimeLeft -= Time.deltaTime;
            if (gameOverMessageTimeLeft < 0f)
            {
                ChangeState(State.SELECTING_PLAYMODE);
            }
        }
    }

    private bool TestGameOverConditionAndSetMessage()
    {
        bool gameOver = false;

        // network error or other player left
        if (OtherPlayerLeft())
        {
            gameOverMessage = "Draw - Network Error or other player left";
            gameOver = true;
        }
        // crane loses when running out of coins
        else if (numCoinsCurrent < NUM_COINS_PER_MOVE)
        {
            gameOverMessage = "Squeaky Snake Wins - Santa lost all his money";
            gameOver = true;
        }
        // snake ate all balls
        else if (GetScoreSnake01() + Mathf.Epsilon >= GetMaxBallsCurrent01())
        {
            // compare scores
            const float DRAW_THRESHOLD = 0.025f;
            float snakeAdvantage = GetScoreSnake01() - GetScoreCrane01();
            if (snakeAdvantage > DRAW_THRESHOLD)
            {
                gameOverMessage = "Squeaky Snake Wins";
            }
            else if (snakeAdvantage < -DRAW_THRESHOLD)
            {
                gameOverMessage = "Santa Claws Wins";
            }
            else
            {
                gameOverMessage = "Draw!";
            }

            gameOver = true;
        }
        // snake isn't there any more (so the crane removed it)
        else if (GameObject.FindObjectOfType<Snake>() == null)
        {
            gameOverMessage = "Santa Claws Wins - Obviously";
            gameOver = true;
        }

        return gameOver;
    }

    private void ChangeState(State newState)
    {
        //Debug.Log("Game.ChangeState called, newState: " + newState);

        // exit old state
        if (state == State.INITIAL)
        {
            craneCamera = GameObject.Find("CraneCamera");
            snakeCamera = GameObject.Find("SnakeCamera");
            titleCamera = GameObject.Find("TitleCamera");

            craneCamera.active = false;
            snakeCamera.active = false;
            titleCamera.active = true;
        }
        else if (state == State.SHOWING_TITLE_SCREEN)
        {
        }
        else if (state == State.INITIALIZE_NETWORK)
        {
        }
        else if (state == State.JOIN_OR_START_ROOM)
        {
        }
        else if (state == State.WAITING_FOR_PLAYER_TWO)
        {
        }
        else if (state == State.LOAD_GAME)
        {
        }
        else if (state == State.TIMER)
        {
        }
        else if (state == State.PLAYING)
        {
        }
        else if (state == State.SHOWING_GAME_OVER_MESSAGE)
        {
            DestroyBallPit();
            DestroyBalls();
            DestroySnake();
            Disconnect();
            // Reset
            playMode = PlayMode.Unselected;
            playerType = PlayerType.INITIAL;
            playerTypeSelected = PlayerType.INITIAL;

            craneCamera.active = false;
            snakeCamera.active = false;
            titleCamera.active = true;
        }

        // init new state
        if (newState == State.INITIAL)
        {
        }
        else if (newState == State.SHOWING_TITLE_SCREEN)
        {
        }
        else if (newState == State.INITIALIZE_NETWORK)
        {
            InitializeNetwork();
        }
        else if (newState == State.JOIN_OR_START_ROOM)
        {
            PlayerType playerTypeOpponent = GetPlayerTypeOpponent(playerType);
            networkInfo.text = "Finding " + playerTypeOpponent + " opponent..";
            GetCustomRoomList(" " + (int)playerTypeOpponent);
        }
        else if (newState == State.LOAD_GAME)
        {
            networkInfo.text = "Preparing game..";
            CreateBallPit();
            CreateSnakeHead();
            CreateBalls();
            CreateCrane();
        }
        else if (newState == State.TIMER)
        {
            networkInfo.text = "3";
            StartTimer();
            MoveCamera();
        }
        else if (newState == State.PLAYING)
        {
            networkInfo.text = "";


            if (playerType == PlayerType.SNAKE)
            {
                Snake snake = GameObject.FindObjectOfType<Snake>();
                snake.EnablePlayerInput();
                snake.SetIsUpdatedByServer(false);

            }
            else if (playerType == PlayerType.CRANE)
            {
                Crane crane = GameObject.FindObjectOfType<Crane>();
                crane.EnablePlayerInput();
                crane.SetIsUpdatedByServer(playMode == PlayMode.Multiplayer);
            }

        }
        else if (newState == State.SHOWING_GAME_OVER_MESSAGE)
        {
            gameOverMessageTimeLeft = GAME_OVER_MESSAGE_DURATION;
        }

        state = newState;
    }


    private float GetMaxBallsCurrent01()
    {
        return 1f - ((numBallsGrabbedByCrane + Snake.NumBallsDestroyedBySnake) / (float)(MAX_BALL_AMOUNT - numBallsDestroyedByLevel));
    }

    private float GetScoreSnake01() // This cannot go over GetMaxBallsCurrent01
    {
        return (float)(numBallsEatenBySnake - Snake.NumBallsDestroyedBySnake) / (float)(MAX_BALL_AMOUNT - numBallsDestroyedByLevel);
    }

    private float GetScoreCrane01()
    {
        return (float)numCoinsCurrent / (float)NUM_COINS_MAX;
    }

    private void ChangePlayerType(PlayerType newPlayerType)
    {
        if (newPlayerType == PlayerType.SNAKE)
        {

        }
        else if (newPlayerType == PlayerType.CRANE)
        {

        }
        playerType = newPlayerType;
    }

    private PlayerType GetPlayerTypeOpponent(PlayerType playerType)
    {
        if (playerType == PlayerType.CRANE)
        {
            return PlayerType.SNAKE;
        }
        if (playerType == PlayerType.SNAKE)
        {
            return PlayerType.CRANE;
        }
        return PlayerType.INITIAL;
    }


    private void StartTimer()
    {
        if (playMode == PlayMode.LocalDebug)
        {
            timerStartTime = Time.time;
        }

        else if (playMode == PlayMode.Multiplayer)
        {
            if (PhotonNetwork.IsMasterClient)
            {
                // Set time
                timerStartTime = PhotonNetwork.Time;
                this.photonView.RPC("SetTimerStarted", RpcTarget.All, timerStartTime);
            }
        }
    }

    private double CalculateElapsedTime()
    {
        if (playMode == PlayMode.LocalDebug)
        {
            return Time.time - timerStartTime;
        }
        else if (playMode == PlayMode.Multiplayer)
        {
            return PhotonNetwork.Time - timerStartTime;
        }
        return 0;
    }


    private void CreateBallPit()
    {
        Instantiate(Resources.Load("Prefabs/Cabinet", typeof(GameObject)));

        GameObject floorCube = Instantiate(Resources.Load("Prefabs/Floorcube", typeof(GameObject))) as GameObject;
        floorCube.transform.localScale = new Vector3(BALL_PIT_WIDTH, BALL_PIT_FLOOR_THICKNESS, BALL_PIT_DEPTH);
        floorCube.transform.position = new Vector3(0f, 0f - BALL_PIT_FLOOR_THICKNESS / 2f, 0f);
        floorCube.transform.name = "Floor";

        GameObject leftWallCube = Instantiate(Resources.Load("Prefabs/WallCube", typeof(GameObject))) as GameObject;
        leftWallCube.transform.localScale = new Vector3(BALL_PIT_WALL_THICKNESS, BALL_PIT_WALL_HEIGHT, BALL_PIT_DEPTH);
        leftWallCube.transform.position = new Vector3(0f - BALL_PIT_WIDTH / 2f - BALL_PIT_WALL_THICKNESS / 2f,
                                                        0f,
                                                        0f);
        leftWallCube.transform.name = "LeftWall";

        GameObject rightWallCube = Instantiate(Resources.Load("Prefabs/WallCube", typeof(GameObject))) as GameObject;
        rightWallCube.transform.localScale = new Vector3(BALL_PIT_WALL_THICKNESS, BALL_PIT_WALL_HEIGHT, BALL_PIT_DEPTH);
        rightWallCube.transform.position = new Vector3(BALL_PIT_WIDTH / 2f + BALL_PIT_WALL_THICKNESS / 2f,
                                                         0f,
                                                         0f);
        rightWallCube.transform.name = "RightWall";

        GameObject farWallCube = Instantiate(Resources.Load("Prefabs/WallCube", typeof(GameObject))) as GameObject;
        farWallCube.transform.localScale = new Vector3(BALL_PIT_WIDTH, BALL_PIT_WALL_HEIGHT, BALL_PIT_WALL_THICKNESS);
        farWallCube.transform.position = new Vector3(0f,
                                                         0f,
                                                         BALL_PIT_DEPTH / 2f + BALL_PIT_WALL_THICKNESS / 2f);
        farWallCube.transform.name = "FarWall";

        GameObject nearWallCube = Instantiate(Resources.Load("Prefabs/WallCube", typeof(GameObject))) as GameObject;
        nearWallCube.transform.localScale = new Vector3(BALL_PIT_WIDTH, BALL_PIT_WALL_HEIGHT, BALL_PIT_WALL_THICKNESS);
        nearWallCube.transform.position = new Vector3(0f,
                                                         0f,
                                                         0f - BALL_PIT_DEPTH / 2f - BALL_PIT_WALL_THICKNESS / 2f);
        nearWallCube.transform.name = "NearWall";
    }

    private void CreateSnakeHead()
    {
        Vector2 minPosition = new Vector2(0f - BALL_PIT_WIDTH / 2f + SNAKE_HEAD_DIAMETER / 2f, BALL_PIT_DEPTH / 2f - SNAKE_HEAD_DIAMETER / 2f);
        Vector2 maxPosition = new Vector2(BALL_PIT_WIDTH / 2f - SNAKE_HEAD_DIAMETER / 2f, 0f - BALL_PIT_DEPTH / 2f + SNAKE_HEAD_DIAMETER / 2f);



        Vector3 snakePosition = new Vector3(Random.Range(minPosition.x, maxPosition.x), SNAKE_HEAD_DIAMETER / 2f, Random.Range(minPosition.y, maxPosition.y)); ;
        if (playMode == PlayMode.LocalDebug)
        {
            Instantiate(Resources.Load("Prefabs/Snake", typeof(GameObject)), snakePosition, Quaternion.identity);
        }
        else if (playMode == PlayMode.Multiplayer)
        {
            PhotonNetwork.InstantiateRoomObject("Prefabs/PhotonSnake", snakePosition, Quaternion.identity, 0, null);
        }
    }

    //TODO Jerry to implement? this needs to return true when there is a network error / other player isn't there any more
    public bool OtherPlayerLeft()
    {
        bool otherPlayerLeft = false;

        //TODO remove this check
        if (Input.GetKeyDown(KeyCode.L))
        {
            otherPlayerLeft = true;
        }

        return otherPlayerLeft;
    }

    private void CreateBalls()
    {
        int ballAmount = Random.Range(MIN_BALL_AMOUNT, MAX_BALL_AMOUNT);

        Vector2 minBallPosition = new Vector2(0f - BALL_PIT_WIDTH / 2f + MAX_BALL_DIAMETER / 2f, BALL_PIT_DEPTH / 2f - MAX_BALL_DIAMETER / 2f);
        Vector2 maxBallPosition = new Vector2(BALL_PIT_WIDTH / 2f - MAX_BALL_DIAMETER / 2f, 0f - BALL_PIT_DEPTH / 2f + MAX_BALL_DIAMETER / 2f);

        float ballY = SNAKE_HEAD_DIAMETER + MAX_BALL_DIAMETER;
        int ballGridX = 0;
        int ballGridY = 0;
        for (int i = 0; i < ballAmount; ++i)
        {

            //NOTE: 8 is just a number I grabbed: put in a constant or calculate
            ++ballGridX;
            if (ballGridX >= 8)
            {
                ballGridX = 0;
                ++ballGridY;

                if (ballGridY >= 8)
                {
                    ballGridY = 0;
                    ballY += MAX_BALL_DIAMETER;
                }
            }

            Vector3 ballPosition = new Vector3(Mathf.Lerp(minBallPosition.x, maxBallPosition.x, ballGridX / 8f),
                                        ballY,
                                        Mathf.Lerp(minBallPosition.y, maxBallPosition.y, ballGridY / 8f));


            if (playMode == PlayMode.LocalDebug)
            {
                GameObject ball = Instantiate(Resources.Load("Prefabs/Ball", typeof(GameObject))) as GameObject;
                ball.transform.name = "Ball" + i;
                ball.transform.position = ballPosition;
            }
            else if (playMode == PlayMode.Multiplayer)
            {
                PhotonNetwork.InstantiateRoomObject("Prefabs/PhotonBall", ballPosition, Quaternion.identity, 0, null);
            }
        }
    }

    private void CreateCrane()
    {
        if (playMode == PlayMode.LocalDebug)
        {
            Instantiate(Resources.Load("Prefabs/Crane", typeof(GameObject)));
        }
        else if (playMode == PlayMode.Multiplayer)
        {
            PhotonNetwork.InstantiateRoomObject("Prefabs/PhotonCrane", Vector3.zero, Quaternion.identity, 0, null);
        }
    }


    private void MoveCamera()
    {
        if (playerType == PlayerType.CRANE)
        {
            craneCamera.active = true;
            snakeCamera.active = false;
            titleCamera.active = false;
        }
        else if (playerType == PlayerType.SNAKE)
        {
            craneCamera.active = false;
            snakeCamera.active = true;
            titleCamera.active = false;
        }
    }

    private void SetBallsEdible()
    {
        Ball[] balls = GameObject.FindObjectsOfType<Ball>();
        int numBallsToTurnEdible = 0;
        {
            int numEdibleBalls = 0;
            foreach (Ball ball in balls)
            {
                if (ball.IsEdible())
                {
                    ++numEdibleBalls;
                }
            }
            numBallsToTurnEdible = NUM_EDIBLES - numEdibleBalls;
        }

        if (numBallsToTurnEdible > 0)
        {
            List<Ball> visibleBalls = new List<Ball>(balls.Length);

            foreach (Ball ball in balls)
            {
                Snake snake = GameObject.FindObjectOfType<Snake>();
                if (!snake.HasBall(ball) && DetermineIfBallIsCloseToFloorAndPropagateToBall(ball))
                {
                    visibleBalls.Add(ball);
                }
            }

            if (visibleBalls.Count > numBallsToTurnEdible)
            {
                int amountOfEdibleBalls = 0;

                while (amountOfEdibleBalls < numBallsToTurnEdible)
                {
                    int randomBallIndex = Random.Range(0, visibleBalls.Count);

                    if (!visibleBalls[randomBallIndex].IsEdible())
                    {
                        visibleBalls[randomBallIndex].SetEdible(true);
                    }
                    ++amountOfEdibleBalls;
                }
            }
            else
            {
                foreach (Ball visibleBall in visibleBalls)
                {
                    visibleBall.SetEdible(true);
                }
            }
        }
    }

    private void DestroyBallPit()
    {
        Destroy(GameObject.Find("Cabinet"));
        Destroy(GameObject.Find("Floor"));
        Destroy(GameObject.Find("LeftWall"));
        Destroy(GameObject.Find("RightWall"));
        Destroy(GameObject.Find("FarWall"));
        Destroy(GameObject.Find("NearWall"));
    }

    private void DestroyBalls()
    {
        foreach (Ball ball in GameObject.FindObjectsOfType<Ball>())
        {
            Destroy(ball.gameObject);
        }
    }

    private void DestroySnake()
    {
        GameObject.Destroy(FindObjectOfType<Snake>().gameObject);
    }

    private void DestroyFallenBalls()
    {
        Snake snake = GameObject.FindObjectOfType<Snake>();
        foreach (Ball ball in GameObject.FindObjectsOfType<Ball>())
        {
            if (ball.transform.position.y < -100f)
            {
                snake.RemoveBallIfPresent(ball);
                Destroy(ball.gameObject);
                ++numBallsDestroyedByLevel;
            }
        }
    }

    private void HideBallsNotOnFloor()
    {
        foreach (Ball ball in GameObject.FindObjectsOfType<Ball>())
        {
            Renderer ballRenderer = ball.GetComponent<Renderer>();

            if (!GameObject.FindObjectOfType<Snake>().HasBall(ball))
            {
                if (DetermineIfBallIsCloseToFloorAndPropagateToBall(ball))
                {
                    float ballAlpha = 1f;
                    ballRenderer.enabled = true;
                    ToOpaqueMode(ballRenderer.material);
                    ballRenderer.material.color = new Color(ballRenderer.material.color.r, ballRenderer.material.color.g, ballRenderer.material.color.b, ballAlpha);
                }
                else if (IsBallSlightlyAboveFloor(ball))
                {
                    ballRenderer.enabled = true;

                    float ballAlpha = GetSlightlyAboveFloorBallAlpha(ball);
                    ToFadeMode(ballRenderer.material);
                    ballRenderer.material.color = new Color(ballRenderer.material.color.r, ballRenderer.material.color.g, ballRenderer.material.color.b, ballAlpha);
                }
                else
                {
                    ballRenderer.enabled = false;
                }
            }
        }
    }

    private bool DetermineIfBallIsCloseToFloorAndPropagateToBall(Ball ball)
    {
        bool closeToFloor = ball.transform.position.y < MAX_BALL_DIAMETER * 0.5;
        ball.SetOnFloor(closeToFloor);
        return closeToFloor;
    }

    private bool IsBallSlightlyAboveFloor(Ball ball)
    {
        bool slightlyAboveFloor = ball.transform.position.y < MAX_BALL_DIAMETER * 1.5f;
        DetermineIfBallIsCloseToFloorAndPropagateToBall(ball);
        return slightlyAboveFloor;
    }

    private float GetSlightlyAboveFloorBallAlpha(Ball ball)
    {
        float t = (ball.transform.position.y - MAX_BALL_DIAMETER * 0.5f) / MAX_BALL_DIAMETER;
        t = -(Mathf.Sin(Mathf.PI + (Mathf.PI * t) / 2f));
        return Mathf.Lerp(1f, 0f, t);
    }


    [PunRPC]
    private void SetTimerStarted( double serverTimeStarted )
    {
        timerStartTime = serverTimeStarted;
    }

    private void InitializeNetwork()
    {
        networkInfo.text = "Enabling network..";
        PhotonNetwork.AutomaticallySyncScene = true;
        PhotonNetwork.ConnectUsingSettings();
        PhotonNetwork.GameVersion = "" + GAME_VERSION;
    }

    private void Disconnect()
    {
        if (playMode == PlayMode.Multiplayer)
        {
            networkInfo.text = "Disconnecting..";
            PhotonNetwork.Disconnect();
        }
    }

    public override void OnConnectedToMaster()
    {
        Debug.Log("PUN Basics Tutorial/Launcher: OnConnectedToMaster() was called by PUN");
        isConnectdToMaster = true;
    }

    private bool IsNetworkInitialized()
    {
        return PhotonNetwork.IsConnectedAndReady && isConnectdToMaster;
    }

    public override void OnDisconnected(DisconnectCause cause)
    {
        networkInfo.text = "Disconnected: " + cause;
        Debug.LogWarningFormat("PUN Basics Tutorial/Launcher: OnDisconnected() was called by PUN with reason {0}", cause);
    }


    private void CreateRoom()
    {
        RoomOptions roomOptions = new RoomOptions();
        roomOptions.CustomRoomProperties = new Hashtable { { PROP_KEY_P1, (int)playerType } };
        roomOptions.CustomRoomPropertiesForLobby = new string[] { PROP_KEY_P1 };
        PhotonNetwork.CreateRoom(null, roomOptions, sqlLobby);
    }

    public override void OnRoomPropertiesUpdate(Hashtable propertiesThatChanged)
    {
        Debug.Log("OnRoomPropertiesUpdate");
    }

    public override void OnCreateRoomFailed(short returnCode, string message)
    {
        Debug.LogErrorFormat("Room creation failed with error code {0} and error message {1}", returnCode, message);
    }


    public override void OnJoinRandomFailed(short returnCode, string message)
    {
        Debug.Log("PUN Basics Tutorial/Launcher:OnJoinRandomFailed() was called by PUN. No random room available, so we create one.\nCalling: PhotonNetwork.CreateRoom");
    }

    public override void OnJoinedRoom()
    {
        Debug.Log("PUN Basics Tutorial/Launcher: OnJoinedRoom() called by PUN. Now this client is in a room.");
        networkInfo.text = "Success..";
        if (PhotonNetwork.CurrentRoom.PlayerCount < 2)
        {
            networkInfo.text = "Waiting for a " + GetPlayerTypeOpponent(playerType) + " opponent..";
        }
    }


    private void GetCustomRoomList(string sqlLobbyFilter)
    {
        isRoomlistUptodate = false;
        PhotonNetwork.GetCustomRoomList(sqlLobby, sqlLobbyFilter);
    }

    public override void OnRoomListUpdate(List<RoomInfo> roomList)
    {
        this.roomList = roomList;
        isRoomlistUptodate = true;
    }   

    private RoomInfo FindRoom()
    {
        foreach (RoomInfo room in roomList)
        {
            if (room.IsOpen && (int)room.CustomProperties[PROP_KEY_P1] != (int)playerTypeSelected)
            {
                networkInfo.text = "Found room '" + room.Name + "'; joining now...";
                return room;
            }
        }
        return null;
    }

    private void OnGUI()
    {
        if (state == State.SELECTING_PLAYMODE)
        {
            GUI.skin.button.alignment = TextAnchor.MiddleCenter;
            GUI.skin.font = Resources.Load<Font>("UI/font");
            GUI.skin.button.fontSize = (int)(Screen.height * 0.03f);

            if (GUI.Button(UIRect(204-60, 185, 680+120, 383), "Multiplayer\n\n(recommended)"))
            {
                playMode = PlayMode.Multiplayer;
            }
            if (GUI.Button(UIRect(204-60, 724, 680+120, 383), "Single\n\nPlayer"))
            {
                playMode = PlayMode.LocalDebug;
            }
        }
        else if (state == State.SHOWING_TITLE_SCREEN)
        {
            GUI.skin.button.alignment = TextAnchor.MiddleCenter;
            GUI.skin.font = Resources.Load<Font>("UI/font");
            GUI.skin.button.fontSize = (int)(Screen.height * 0.03f);

            if (GUI.Button(UIRect(204-60, 185, 680+120, 383), "Play as Squeaky Snake"))
            {
                playerTypeSelected = PlayerType.SNAKE;
            }
            if (GUI.Button(UIRect(204-60, 724, 680+120, 383), "Play as Santa Claw"))
            {
                playerTypeSelected = PlayerType.CRANE;
            }
        }
        else if (state == State.JOIN_OR_START_ROOM)
        {
        }
        else if (state == State.WAITING_FOR_PLAYER_TWO)
        {
        }
        else if (state == State.PLAYING)
        {
            OnGUI_IngameUI();
        }
        else if (state == State.SHOWING_GAME_OVER_MESSAGE)
        {
            OnGUI_IngameUI();

            GUI.skin.label.alignment = TextAnchor.MiddleCenter;
            GUI.skin.font = Resources.Load<Font>("UI/font");
            GUI.skin.label.fontSize = (int)(Screen.height * 0.05f);
            OutlinedLabel(UIRect(0, 0, 3000, 2000), gameOverMessage);
        }
    }


    private void OnGUI_IngameUI()
    {
        //backup various values here
        {
            // left side
            GUI.DrawTexture(UIRect(123, 35, 200, 200), Resources.Load<Texture>("UI/snakeIcon"));
            GUI.DrawTexture(UIRect(123, 345, 120, 120), Resources.Load<Texture>("UI/ballsIcon"));
            GUI.DrawTexture(UIRect(123, 253, 1035, 74), Resources.Load<Texture>("UI/emptyProgressBar"));

            float snakeBarCapacity = GetMaxBallsCurrent01();
            float snakeProgress = GetScoreSnake01();
            GUI.DrawTextureWithTexCoords(UIRect(123, 253, (int)(1035f * snakeProgress), 74), Resources.Load<Texture>("UI/fullProgressBar"), new Rect(0f, 0f, snakeProgress, 1f));
            GUI.DrawTextureWithTexCoords(UIRect(123 + (int)(1035f * snakeBarCapacity), 253, 1035 - (int)(1035f * snakeBarCapacity), 74), Resources.Load<Texture>("UI/reducedProgressBar"), new Rect(snakeBarCapacity, 0f, 1f, 1f));


            GUI.skin.label.alignment = TextAnchor.MiddleLeft;
            GUI.skin.font = Resources.Load<Font>("UI/font");
            GUI.skin.label.fontSize = (int)(Screen.height * 0.05f);
            OutlinedLabel(UIRect(359, 35, 1035, 200), "Squeaky Snake");
            GUI.skin.label.fontSize = (int)(Screen.height * 0.03f);
            OutlinedLabel(UIRect(261, 345, 1035, 120), "" + numBallsEatenBySnake + " / " + MIN_BALL_AMOUNT);

            // center
            GUI.DrawTexture(UIRect(1500 - 180 / 2, 35, 180, 180), Resources.Load<Texture>("UI/VSIcon"));

            // right side
            GUI.DrawTexture(UIRectRightOrigin(123, 35, 200, 200), Resources.Load<Texture>("UI/clawIcon"));
            GUI.DrawTexture(UIRectRightOrigin(123, 345, 120, 120), Resources.Load<Texture>("UI/coinsIcon"));
            GUI.DrawTexture(UIRectRightOrigin(123, 253, 1035, 74), Resources.Load<Texture>("UI/emptyProgressBar"));
            //TODO repeat 10 times (or tunable number, depending on how many coins you have
            //GUI.DrawTexture(UIRectRightOrigin(200, 253, 16, 74), Resources.Load<Texture>("UI/progressBarDivider"));
            float craneProgress = GetScoreCrane01();
            GUI.DrawTextureWithTexCoords(UIRectRightOrigin(123, 253, (int)(1035f * craneProgress), 74), Resources.Load<Texture>("UI/fullProgressBar"), new Rect(0f, 0f, craneProgress, 1f));

            GUI.skin.label.alignment = TextAnchor.MiddleRight;
            GUI.skin.font = Resources.Load<Font>("UI/font");
            GUI.skin.label.fontSize = (int)(Screen.height * 0.05f);
            OutlinedLabel(UIRectRightOrigin(359, 35, 1035, 200), "Santa Claws");
            GUI.skin.label.fontSize = (int)(Screen.height * 0.03f);

            string totalCents = "" + numCoinsCurrent;
            OutlinedLabel(UIRectRightOrigin(261, 345, 1035, 120), "" + numCoinsCurrent / 100 + "." + totalCents.Substring(totalCents.Length - 3, 2));

        }
        //restore various values here
    }

    //input coordinates are in the space of a 3000x2000 px image
    private Rect UIRect(int x, int y, int width, int height)
    {
        Rect rect = new Rect(    1f * x / 3000f * Screen.width,
                            1f * y / 2000f * Screen.height,
                            1f * width / 3000f * Screen.width,
                            1f * height / 2000 * Screen.height);

        return rect;
    }

    //input coordinates are in the space of a 3000x2000 px image
    private Rect UIRectRightOrigin(int x, int y, int width, int height)
    {
        x = 3000 - x - width;
        return UIRect(x, y, width, height);
    }

    private void OutlinedLabel(Rect rect, string text)
    {
        //1 was not enough, 2 and 4 not enough

        GUI.contentColor = Color.black;
        GUI.Label(new Rect(rect.x - Mathf.Max(8f / 2000f * Screen.width, 2f), rect.y, rect.width, rect.height), text);
        GUI.Label(new Rect(rect.x + Mathf.Max(8f / 2000f * Screen.width, 2f), rect.y, rect.width, rect.height), text); //was 4 instead of 100
        GUI.Label(new Rect(rect.x, rect.y + Mathf.Max(8f / 2000f * Screen.height, 2f), rect.width, rect.height), text);
        GUI.Label(new Rect(rect.x, rect.y - Mathf.Max(8f / 2000f * Screen.height, 2f), rect.width, rect.height), text);

        GUI.contentColor = Color.white;
        GUI.Label(rect, text);
    }

    private void ToOpaqueMode( Material material)
    {
        material.SetOverrideTag("RenderType", "");
        material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.One);
        material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.Zero);
        material.SetInt("_ZWrite", 1);
        material.DisableKeyword("_ALPHATEST_ON");
        material.DisableKeyword("_ALPHABLEND_ON");
        material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
        material.renderQueue = -1;
    }

    private void ToFadeMode(Material material)
    {
        material.SetOverrideTag("RenderType", "Transparent");
        material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.SrcAlpha);
        material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
        material.SetInt("_ZWrite", 0);
        material.DisableKeyword("_ALPHATEST_ON");
        material.EnableKeyword("_ALPHABLEND_ON");
        material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
        material.renderQueue = (int)UnityEngine.Rendering.RenderQueue.Transparent;
    }

    private void OnApplicationQuit()
    {
        PhotonNetwork.Disconnect();
    }
}


