﻿using System.Collections;
using System.Collections.Generic;
using Photon.Pun;
using UnityEngine;


[RequireComponent(typeof(Snake), typeof(PhotonView))]
public class SnakePhoton : MonoBehaviourPun, IPunObservable
{
    private Vector3 hostPosition = Vector3.zero;
    private Vector3 hostForward = Vector3.forward;
    private Vector3 hostLocalScale = Vector3.one;
    private bool isUpdateRequired = false;

    public virtual void Awake()
    {
        if (!photonView.IsMine)
        {
            GetComponent<Rigidbody>().isKinematic = true;
        }
    }

    public void OnPhotonSerializeView(PhotonStream stream, PhotonMessageInfo info)
    {
        if (stream.IsWriting)
        {
            //We own this player: send the others our data
            stream.SendNext(transform.position);
            stream.SendNext(transform.forward);
            stream.SendNext(transform.localScale);
        }
        else
        {
            //Network player, receive data
            isUpdateRequired = true;
            hostPosition = (Vector3)stream.ReceiveNext();
            hostForward = (Vector3)stream.ReceiveNext();
            hostLocalScale = (Vector3)stream.ReceiveNext();
        }
    }

    // Update is called once per frame
    void Update()
    {
        if (!photonView.IsMine && isUpdateRequired)
        {
            //Update Object position and Rigidbody parameters
            transform.position = Vector3.Lerp(transform.position, hostPosition, Time.deltaTime * 5);
            transform.forward = Vector3.Lerp(transform.forward, hostForward, Time.deltaTime * 5);
            transform.localScale = Vector3.Lerp(transform.localScale, hostLocalScale, Time.deltaTime * 5);
        }
    }
}
